from consts import *
import Utils
import Live

class SceneDataHandler:
    """
    Handles request for scene data on the assigned channel via midi cc messages
    """
    def __init__(self,channel,controller,logger):
        self.controller = controller
        self.logger = logger
        self.channel = channel
        self.builder = SceneSysexBuilder()
    
    def build_midi_map(self,script_handle,midi_map_handle):
        map_mode = Live.MidiMap.MapMode.absolute
        Live.MidiMap.forward_midi_cc(script_handle,midi_map_handle,self.channel,SCENE_DATA_CC)
    
    def handle_cc(self,chan,num,val):
        """
        Only need to use the value of the cc which is the scene index.
        """
        self.logger.log("SDH " + str(num) + "," + str(val))
        return self.builder.getSceneData(val)
    
    def can_handle(self,chan,num,val):
        return chan == self.channel and num == SCENE_DATA_CC
        
class SceneMonitorHandler:
    """
    Handles requests for listening to scene attributes
    """
    def __init__(self, channel, controller,logger):
        self.controller = controller
        self.logger = logger
        self.channel = channel
        self.monitors = []
        # Create storage for our listeners
        for index in range(128):
            self.monitors.append(None)
        
    def build_midi_map(self,script_handle,midi_map_handle):
        map_mode = Live.MidiMap.MapMode.absolute
        Live.MidiMap.forward_midi_cc(script_handle,midi_map_handle,self.channel,SCENE_MONITOR_CC)
        Live.MidiMap.forward_midi_cc(script_handle,midi_map_handle,self.channel,SCENE_STOP_MONITOR_CC)

    def handle_cc(self,chan,num,val):
        sceneMonitor = self.monitors[val]

        # If the monitor doesn't exist, create it
        if not sceneMonitor:
            self.logger.log("Creating scene monitor at " + str(val))
            sceneMonitor = SceneMonitor(self.controller,self.logger)
            self.monitors[val] = sceneMonitor

        if num == SCENE_MONITOR_CC:
            if sceneMonitor.isActive():
                sceneMonitor.stop()
            sceneMonitor.start(val)
        elif num == SCENE_STOP_MONITOR_CC:
            sceneMonitor.stop()

    def can_handle(self,chan,num,val):
        canHandleCC = (num == SCENE_MONITOR_CC) or (num == SCENE_STOP_MONITOR_CC)
        return chan == self.channel and canHandleCC

class SceneMonitor:
    """
    Monitors a single scene for changes
    """
    def __init__(self,controller,logger):
        self.controller = controller
        self.logger = logger
        self.logger.log("New SceneMonitor")
        self.builder = SceneSysexBuilder()
        self.active = 0
        
    def start(self,sceneIndex):
        self.sceneIndex = sceneIndex
        self.scene = Utils.getScene(self.sceneIndex)
        self.sceneName = self.scene.name
        
        if self.scene:
            self.logger.log("Monitoring Scene " + str(self.sceneIndex))
            self.listenToScene()
            self.active = 1

    def stop(self):
        self.logger.log("Stop Monitoring Scene " + str(self.sceneIndex))
        self.stopListeningToScene()
        self.active = 0
                   
    def listenToScene(self):
        if self.scene:
            if self.scene.name_has_listener(self.nameChanged):
                self.logger.log("Scene already has listener "  + str(self.sceneIndex) + " : " + self.scene.name)
            else:
                self.logger.log("Listening to scene " + str(self.sceneIndex) + " : " + self.scene.name)
                self.scene.add_name_listener(self.nameChanged)
            self.sceneChanged()
        else:
            self.logger.log("No scene to monitor " + str(self.sceneIndex))
        
    def stopListeningToScene(self):
        if self.scene:
            self.logger.log("Stopping listening to scene " + str(self.sceneIndex) + " : " + self.scene.name)
            self.scene.remove_name_listener(self.nameChanged)

    def isActive(self):
        return self.active
    
    def sceneChanged(self):
        """
        Generic scene change handler
        """
        self.logger.log("Scene changed " + str(self.sceneIndex))
        self.controller.sendSysex(self.builder.getSceneData(self.sceneIndex))
    
    def nameChanged(self):
        if self.scene.name != self.sceneName:
            self.logger.log("Scene name changed : " + self.scene.name + " from " + self.sceneName)
            self.sceneName = self.scene.name
            data = SCENE_NAME + (self.sceneIndex,) + Utils.translateString(self.scene.name)
            self.controller.sendSysex(data)
        else:
            self.logger.log("Scene nameChanged invoked but name is the same")
        

class SceneTriggerHandler:
    """
    Handles requests to fire scenes
    """
    def __init__(self, channel, controller, logger):
        self.controller = controller
        self.logger = logger
        self.channel = channel
    
    def build_midi_map(self,script_handle,midi_map_handle):
        map_mode = Live.MidiMap.MapMode.absolute
        Live.MidiMap.forward_midi_cc(script_handle,midi_map_handle,self.channel,FIRE_SCENE_CC)
    
    def handle_cc(self,chan,num,val):
        self.logger.log("Trigger scene " + str(val))
        scene = Utils.getScene(val)
        if scene:
            scene.fire()

    def can_handle(self,chan,num,val):
        return chan == self.channel and num == FIRE_SCENE_CC

class SceneSysexBuilder:
    def getSceneData(self,sceneIndex):
        scene = Utils.getScene(sceneIndex)
        
        if scene:
            return self.buildSceneData(sceneIndex,scene)
        else:
            return NO_SCENE + (sceneIndex,)
        
    def buildSceneData(self,index,scene):
        nameData = Utils.translateString(scene.name)
        data = SCENE_DATA + (index,) + nameData
        return data
    